from __future__ import absolute_import, print_function, unicode_literals

import time

import dill
import mock
import pytest

from kq import Job, Queue

from .utils import (
    success_func,
    failure_func
)


@pytest.fixture(autouse=True)
def producer(monkeypatch):
    mock_producer_inst = mock.MagicMock()
    mock_producer_cls = mock.MagicMock()
    mock_producer_cls.return_value = mock_producer_inst
    monkeypatch.setattr('kafka.KafkaProducer', mock_producer_cls)
    return mock_producer_cls, mock_producer_inst


@pytest.fixture(autouse=True)
def logger(monkeypatch):
    mock_logger_inst = mock.MagicMock()
    mock_get_logger = mock.MagicMock()
    mock_get_logger.return_value = mock_logger_inst
    monkeypatch.setattr('logging.getLogger', mock_get_logger)
    return mock_logger_inst


def test_init(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(
        hosts='host:7000,host:8000',
        topic='foo',
        timeout=1000,
        compression='gzip',
        acks=0,
        retries=5,
        job_size=10000000,
        cafile='/test/files/cafile',
        certfile='/test/files/certfile',
        keyfile='/test/files/keyfile',
        crlfile='/test/files/crlfile'
    )
    producer_cls.assert_called_with(
        bootstrap_servers='host:7000,host:8000',
        compression_type='gzip',
        acks=0,
        retries=5,
        max_request_size=10000000,
        buffer_memory=33554432,
        ssl_cafile='/test/files/cafile',
        ssl_certfile='/test/files/certfile',
        ssl_keyfile='/test/files/keyfile',
        ssl_crlfile='/test/files/crlfile',
    )
    assert repr(queue) == 'Queue(topic=foo)'
    assert queue.hosts == ['host:7000', 'host:8000']
    assert queue.timeout == 1000
    assert queue.topic == 'foo'
    assert queue.producer == producer_inst
    assert not logger.info.called


def test_enqueue_call(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo', timeout=300)
    job = queue.enqueue(success_func, 1, 2, c=[3, 4, 5])

    assert isinstance(job, Job)
    assert isinstance(job.id, str)
    assert isinstance(job.timestamp, int)
    assert job.topic == 'foo'
    assert job.func == success_func
    assert job.args == (1, 2)
    assert job.kwargs == {'c': [3, 4, 5]}
    assert job.timeout == 300

    producer_inst.send.assert_called_with('foo', dill.dumps(job), key=None)
    logger.info.assert_called_once_with('Enqueued: {}'.format(job))


def test_enqueue_call_with_key(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo', timeout=300)
    job = queue.enqueue_with_key('bar', success_func, 1, 2, c=[3, 4, 5])

    assert isinstance(job, Job)
    assert isinstance(job.id, str)
    assert isinstance(job.timestamp, int)
    assert job.topic == 'foo'
    assert job.func == success_func
    assert job.args == (1, 2)
    assert job.kwargs == {'c': [3, 4, 5]}
    assert job.timeout == 300
    assert job.key == 'bar'

    producer_inst.send.assert_called_with('foo', dill.dumps(job), key='bar')
    logger.info.assert_called_once_with('Enqueued: {}'.format(job))


def test_invalid_call(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo', timeout=300)

    for bad_func in [None, 1, {1, 2}, [1, 2, 3]]:
        with pytest.raises(ValueError) as e:
            queue.enqueue(bad_func, 1, 2, a=3)
        assert str(e.value) == '{} is not a callable'.format(bad_func)

    assert not producer_inst.send.called
    assert not logger.info.called


def test_invalid_call_with_key(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo', timeout=300)

    for bad_func in [None, 1, {1, 2}, [1, 2, 3]]:
        with pytest.raises(ValueError) as e:
            queue.enqueue_with_key('foo', bad_func, 1, 2, a=3)
        assert str(e.value) == '{} is not a callable'.format(bad_func)

    assert not producer_inst.send.called
    assert not logger.info.called


def test_enqueue_job(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo', timeout=300)

    old_job = Job(
        id='2938401',
        timestamp=int(time.time()),
        topic='bar',
        func=failure_func,
        args=[1, 2],
        kwargs={'a': 3},
        timeout=100,
    )
    new_job = queue.enqueue(old_job)

    assert isinstance(new_job, Job)
    assert isinstance(new_job.id, str)
    assert isinstance(new_job.timestamp, int)
    assert old_job.id != new_job.id
    assert old_job.timestamp <= new_job.timestamp
    assert new_job.topic == 'foo'
    assert new_job.func == failure_func
    assert new_job.args == [1, 2]
    assert new_job.kwargs == {'a': 3}
    assert new_job.timeout == 300
    assert new_job.key == None

    producer_inst.send.assert_called_with(
        'foo', dill.dumps(new_job), key=None
    )
    logger.info.assert_called_once_with('Enqueued: {}'.format(new_job))


def test_enqueue_job_with_key(producer, logger):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo', timeout=300)

    old_job = Job(
        id='2938401',
        timestamp=int(time.time()),
        topic='bar',
        func=failure_func,
        args=[1, 2],
        kwargs={'a': 3},
        timeout=100,
        key='bar',
    )
    new_job = queue.enqueue_with_key('baz', old_job)

    assert isinstance(new_job, Job)
    assert isinstance(new_job.id, str)
    assert isinstance(new_job.timestamp, int)
    assert old_job.id != new_job.id
    assert old_job.timestamp <= new_job.timestamp
    assert new_job.topic == 'foo'
    assert new_job.func == failure_func
    assert new_job.args == [1, 2]
    assert new_job.kwargs == {'a': 3}
    assert new_job.timeout == 300
    assert new_job.key == 'baz'

    producer_inst.send.assert_called_with(
        'foo', dill.dumps(new_job), key='baz'
    )
    logger.info.assert_called_once_with('Enqueued: {}'.format(new_job))


def test_job_decorator():
    queue = Queue(hosts='host:7000', topic='foo')

    @queue.job
    def test_function(a, b, c=None):
        return a, b, c
    assert hasattr(test_function, 'delay')

    with pytest.raises(Exception) as e:
        test_function.delay(1, 2, 3, 4)
    assert "Can't pickle" in str(e.value)


def test_flush(producer):
    producer_cls, producer_inst = producer

    queue = Queue(hosts='host:7000', topic='foo')
    queue.flush()
    producer_inst.flush.assert_called_once()




from __future__ import absolute_import, print_function, unicode_literals

import functools
import logging
import time
import uuid

import dill
import kafka

from kq.job import Job


class Queue(object):
    """KQ queue.

    A queue serializes incoming function calls and places them into a Kafka
    topic as *jobs*. Workers fetch these jobs and execute them asynchronously
    in the background. Here is an example of initializing and using a queue:

    .. code-block:: python

        from kq import Queue, Job

        queue = Queue(
            hosts='host:7000,host:8000',
            topic='foo',
            timeout=3600,
            compression='gzip',
            acks=0,
            retries=5,
            job_size=10000000,
            cafile='/my/files/cafile',
            certfile='/my/files/certfile',
            keyfile='/my/files/keyfile',
            crlfile='/my/files/crlfile'
        )
        job = queue.enqueue(my_func, *args, **kwargs)
        assert isinstance(job, Job)

    .. note::

        The number of partitions in a Kafka topic limits how many workers
        can read from the queue in parallel. For example, maximum of 10
        workers can work off a queue with 10 partitions.


    :param hosts: Comma-separated Kafka hostnames and ports. For example,
        ``"localhost:9000,localhost:8000,192.168.1.1:7000"`` is a valid input
        string. Default: ``"127.0.0.1:9092"``.
    :type hosts: str | unicode
    :param topic: Name of the Kafka topic. Default: ``"default"``.
    :type topic: str | unicode
    :param timeout: Default job timeout threshold in seconds. If not set, the
        enqueued jobs are left to run until they finish. This means a hanging
        job can potentially block the workers. Default: ``None`` (no timeout).
    :type timeout: int
    :param compression: The algorithm used for compressing job data. Allowed
        values are: ``"gzip"``, ``"snappy"`` and ``"lz4"``. Default: ``None``
        (no compression).
    :type compression: str | unicode
    :param acks: The number of acknowledgements required from the broker(s)
        before considering a job successfully enqueued. Allowed values are:

        .. code-block:: none

            0: Do not wait for any acknowledgment from the broker leader
               and consider the job enqueued as soon as it is added to
               the socket buffer. Persistence is not guaranteed on broker
               failures.

            1: Wait for the job to be saved on the broker leader but not
               for it be replicated across other brokers. If the leader
               broker fails before the replication finishes the job may
               not be persisted.

           -1: Wait for the job to be replicated across all brokers. As
               long as one of the brokers is functional job persistence
               is guaranteed.

        Default: ``1``.

    :type acks: int
    :param retries: The maximum number of attempts to re-enqueue a job when
        the job fails to reach the broker. Retries may alter the sequence of
        the enqueued jobs. Default: ``0``.
    :type retries: int
    :param job_size: The max size of each job in bytes. Default: ``1048576``.
    :type job_size: int
    :param cafile: Full path to the trusted CA certificate file.
    :type cafile: str | unicode
    :param certfile: Full path to the client certificate file.
    :type certfile: str | unicode
    :param keyfile: Full path to the client private key file.
    :type keyfile: str | unicode
    :param crlfile: Full path to the CRL file for validating certification
        expiry. This option is only available with Python 3.4+ or 2.7.9+.
    :type crlfile: str | unicode
    """

    def __init__(self,
                 hosts='127.0.0.1:9092',
                 topic='default',
                 timeout=None,
                 compression=None,
                 acks=1,
                 retries=0,
                 job_size=1048576,
                 cafile=None,
                 certfile=None,
                 keyfile=None,
                 crlfile=None):
        self._hosts = hosts
        self._topic = topic
        self._timeout = timeout
        self._logger = logging.getLogger('kq')
        self._producer = kafka.KafkaProducer(
            bootstrap_servers=self._hosts,
            compression_type=compression,
            acks=acks,
            retries=retries,
            max_request_size=job_size,
            buffer_memory=max(job_size, 33554432),
            ssl_cafile=cafile,
            ssl_certfile=certfile,
            ssl_keyfile=keyfile,
            ssl_crlfile=crlfile
        )

    def __repr__(self):
        """Return a string representation of the queue.

        :return: String representation of the queue.
        :rtype: str | unicode
        """
        return 'Queue(topic={})'.format(self._topic)

    @property
    def producer(self):
        """Return the Kafka producer object.

        :return: Kafka producer object.
        :rtype: kafka.producer.KafkaProducer
        """
        return self._producer

    @property
    def hosts(self):
        """Return the list of Kafka host names and ports.

        :return: List of Kafka host names and ports.
        :rtype: [str]
        """
        return self._hosts.split(',')

    @property
    def topic(self):
        """Return the name of the Kafka topic in use.

        :return: Name of the Kafka topic in use.
        :rtype: str | unicode
        """
        return self._topic

    @property
    def timeout(self):
        """Return the job timeout threshold in seconds.

        :return: Job timeout threshold in seconds.
        :rtype: int
        """
        return self._timeout

    def enqueue(self, obj, *args, **kwargs):
        """Place the function call (or the job) in the Kafka topic.

        For example:

        .. code-block:: python

            import requests
            from kq import Queue

            q = Queue()

            # You can queue the function call with its arguments
            job = q.enqueue(requests.get, 'https://www.google.com')

            # Or you can queue a kq.job.Job instance directly
            q.enqueue(job)

        :param obj: Function or the job object to enqueue. If a function is
            given, the function *must* be pickle-able.
        :type obj: callable | kq.job.Job
        :param args: Arguments for the function. Ignored if a KQ job object
            is given for the first argument instead.
        :type args: list
        :param kwargs: Keyword arguments for the function. Ignored if a KQ
            job instance is given as the first argument instead.
        :type kwargs: dict
        :return: The job that was enqueued
        :rtype: kq.job.Job
        """
        if isinstance(obj, Job):
            func = obj.func
            args = obj.args
            kwargs = obj.kwargs
            key = obj.key
        else:
            func = obj
            key = None

        if not callable(func):
            raise ValueError('{} is not a callable'.format(func))

        job = Job(
            id=str(uuid.uuid4()),
            timestamp=int(time.time()),
            topic=self._topic,
            func=func,
            args=args,
            kwargs=kwargs,
            timeout=self._timeout,
            key=key
        )
        self._producer.send(self._topic, dill.dumps(job), key=key)
        self._logger.info('Enqueued: {}'.format(job))
        return job

    def enqueue_with_key(self, key, obj, *args, **kwargs):
        """Place the function call (or the job) in the Kafka topic with key.

        For example:

        .. code-block:: python

            import requests
            from kq import Queue

            q = Queue()

            url = 'https://www.google.com'

            # You can queue the function call with its arguments
            job = q.enqueue_with_key('my_key', requests.get, url)

            # Or you can queue a kq.job.Job instance directly
            q.enqueue_with_key('my_key', job)

        :param key: The key for the Kafka message. Jobs with the same key are
            guaranteed to be placed in the same Kafka partition and processed
            sequentially. If a job object is enqueued, its key is overwritten.
        :type key: str
        :param obj: Function or the job object to enqueue. If a function is
            given, the function *must* be pickle-able.
        :type obj: callable | kq.job.Job
        :param args: Arguments for the function. Ignored if a KQ job object
            is given for the first argument instead.
        :type args: list
        :param kwargs: Keyword arguments for the function. Ignored if a KQ
            job instance is given as the first argument instead.
        :type kwargs: dict
        :return: The job that was enqueued
        :rtype: kq.job.Job
        """
        if isinstance(obj, Job):
            func = obj.func
            args = obj.args
            kwargs = obj.kwargs
        else:
            func = obj

        if not callable(func):
            raise ValueError('{} is not a callable'.format(func))

        job = Job(
            id=str(uuid.uuid4()),
            timestamp=int(time.time()),
            topic=self._topic,
            func=func,
            args=args,
            kwargs=kwargs,
            timeout=self._timeout,
            key=key
        )
        self._producer.send(self._topic, dill.dumps(job), key=key)
        self._logger.info('Enqueued: {}'.format(job))
        return job

    def job(self, func):
        """Decorator which add a **delay** method to a function.

        When the **delay** method is called, the function is queued as a job.
        For example:

        .. code-block:: python

            from kq import Queue
            queue = Queue()

            @queue.job
            def calculate_sum(a, b, c):
                return a + b + c

            # Enqueue the function as a job
            calculate_sum.delay(1, 2, 3)

        :param func: The function to decorate.
        :type func: callable
        :return: The decorated function with new method **delay**
        :rtype: callable
        """
        @functools.wraps(func)
        def delay(*args, **kwargs):  # pragma: no cover
            return self.enqueue(func, *args, **kwargs)
        func.delay = delay
        return func

    def flush(self):
        """Force-flush all buffered records to the broker."""
        self._logger.info('Flushing {} ...'.format(self))
        self._producer.flush()



from __future__ import absolute_import, print_function, unicode_literals

import multiprocessing
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import sys


class CaptureOutput(list):

    def __enter__(self):
        self._orig_stdout = sys.stdout
        self._temp_stdout = StringIO()
        sys.stdout = self._temp_stdout
        return self

    def __exit__(self, *args):
        self.append(self._temp_stdout.getvalue())
        sys.stdout = self._orig_stdout


def success_func(a, b, c=None):
    return a, b, c


def failure_func(*_):
    raise ValueError('failed!')


def timeout_func(*_):
    raise multiprocessing.TimeoutError



from __future__ import absolute_import, print_function, unicode_literals

from collections import namedtuple


# Namedtuple which encapsulates a KQ job
Job = namedtuple(
    typename='Job',
    field_names=[
        'id',         # UUID of the job
        'timestamp',  # Unix timestamp indicating when the job was queued
        'topic',      # Name of the Kafka topic the job was enqueued in
        'func',       # Job function/callable
        'args',       # Job function arguments
        'kwargs',     # Job function keyword arguments
        'timeout',    # Job timeout threshold in seconds
        'key'         # Jobs w/ the same keys end up in the same partition
    ]
)
Job.__new__.__defaults__ = (None,)



from __future__ import absolute_import, print_function, unicode_literals

from kq.job import Job
from kq.utils import func_repr
from kq.version import VERSION


def test_version():
    assert isinstance(VERSION, str)
    assert all(n.isdigit() for n in VERSION.split('.'))


def test_job():
    job = Job(1, 2, 3, 4, 5, 6, 7, 8)

    assert job.id == 1
    assert job.timestamp == 2
    assert job.topic == 3
    assert job.func == 4
    assert job.args == 5
    assert job.kwargs == 6
    assert job.timeout == 7
    assert job.key == 8


def test_func_repr():

    def f(a, b):
        return a + b

    expected = 'tests.test_misc.f()'
    assert func_repr(f, [], {}) == expected

    expected = 'tests.test_misc.f(1)'
    assert func_repr(f, [1], {}) == expected

    expected = 'tests.test_misc.f(1, 2)'
    assert func_repr(f, [1, 2], {}) == expected

    expected = 'tests.test_misc.f(a=1)'
    assert func_repr(f, [], {'a': 1}) == expected

    expected = 'tests.test_misc.f(a=1, b=2)'
    assert func_repr(f, [], {'a': 1, 'b': 2}) == expected

    expected = 'tests.test_misc.f(1, a=1)'
    assert func_repr(f, [1], {'a': 1}) == expected

    expected = 'tests.test_misc.f(1, 2, a=1)'
    assert func_repr(f, [1, 2], {'a': 1}) == expected

    expected = 'tests.test_misc.f(1, 2, a=1, b=2)'
    assert func_repr(f, [1, 2], {'a': 1, 'b': 2}) == expected



from __future__ import absolute_import, print_function, unicode_literals

import mock
import pytest

from kq import Manager

from .utils import CaptureOutput

expected_info_output = """
Offsets per Topic:

Topic foo:

    Partition 1  : 100
    Partition 2  : 100
    Partition 3  : 100

Topic bar:

    Partition 1  : 100
    Partition 2  : 100
    Partition 3  : 100

Topic baz:

    Partition 1  : 100
    Partition 2  : 100
    Partition 3  : 100
"""


@pytest.fixture(autouse=True)
def consumer(monkeypatch):
    mock_consumer_inst = mock.MagicMock()
    mock_consumer_inst.topics.return_value = ['foo', 'bar', 'baz']
    mock_consumer_inst.partitions_for_topic.return_value = [1, 2, 3]
    mock_consumer_inst.position.return_value = 100
    mock_consumer_cls = mock.MagicMock()
    mock_consumer_cls.return_value = mock_consumer_inst
    monkeypatch.setattr('kafka.KafkaConsumer', mock_consumer_cls)
    return mock_consumer_cls, mock_consumer_inst


def test_init(consumer):
    consumer_cls, consumer_inst = consumer

    manager = Manager(
        hosts='host:7000,host:8000',
        cafile='/test/files/cafile',
        certfile='/test/files/certfile',
        keyfile='/test/files/keyfile',
        crlfile='/test/files/crlfile'
    )
    consumer_cls.assert_called_with(
        bootstrap_servers='host:7000,host:8000',
        ssl_cafile='/test/files/cafile',
        ssl_certfile='/test/files/certfile',
        ssl_keyfile='/test/files/keyfile',
        ssl_crlfile='/test/files/crlfile',
        consumer_timeout_ms=-1,
        enable_auto_commit=True,
        auto_offset_reset='latest',
    )
    assert repr(manager) == 'Manager(hosts=host:7000,host:8000)'
    assert manager.hosts == ['host:7000', 'host:8000']
    assert manager.consumer == consumer_inst


def test_info():
    manager = Manager(
        hosts='host:7000,host:8000',
        cafile='/test/files/cafile',
        certfile='/test/files/certfile',
        keyfile='/test/files/keyfile',
        crlfile='/test/files/crlfile'
    )
    with CaptureOutput() as output:
        manager.info()
    assert len(output) == 1
    assert '\n' + output[0] == expected_info_output



from __future__ import absolute_import, print_function, unicode_literals

import collections
import logging
import multiprocessing as mp
import traceback as tb

import dill
import kafka

from kq.job import Job
from kq.utils import func_repr, rec_repr


class Worker(object):
    """KQ worker.

    A worker fetches jobs from a Kafka broker, de-serializes them and
    executes them asynchronously in the background. Here is an example
    of initializing and starting a worker:

    .. code-block:: python

        from kq import Worker

        worker = Worker(
            hosts='host:7000,host:8000',
            topic='foo',
            timeout=3600,
            callback=None,
            job_size=10000000,
            cafile='/my/files/cafile',
            certfile='/my/files/certfile',
            keyfile='/my/files/keyfile',
            crlfile='/my/files/crlfile',
            proc_ttl=2000,
        )
        worker.start()

    .. note::

        The number of partitions in a Kafka topic limits how many workers
        can read from the queue in parallel. For example, maximum of 10
        workers can work off a queue with 10 partitions.

    :param hosts: Comma-separated Kafka hostnames and ports. For example,
        ``"localhost:9000,localhost:8000,192.168.1.1:7000"`` is a valid
        input string. Default: ``"127.0.0.1:9092"``.
    :type hosts: str | unicode
    :param topic: Name of the Kafka topic. Default: ``"default"``.
    :type topic: str | unicode
    :param timeout: Default job timeout threshold in seconds. If not set, the
        enqueued jobs are left to run until they finish. This means a hanging
        job can potentially block the workers. Default: ``None`` (no timeout).
        If set, overrides timeouts set when the jobs were first enqueued.
    :type timeout: int
    :param callback: Function executed after a job is fetched and processed.
        Its signature must be ``callback(status, job, exception, traceback)``
        where:

        .. code-block:: none

            status (str | unicode)
                The status of the job execution, which can be "timeout",
                "failure" or "success".

            job (kq.job.Job)
                The job namedtuple object consumed by the worker.

            result (object)
                The result of the job execution.

            exception (Exception | None)
                The exception raised while the job was running, or None
                if there were no errors.

            traceback (str | unicode | None)
                The traceback of the exception raised while the job was
                running, or None if there were no errors.

    :type callback: callable
    :param job_size: The max size of each job in bytes. Default: ``1048576``.
    :type job_size: int
    :param cafile: Full path to the trusted CA certificate file.
    :type cafile: str | unicode
    :param certfile: Full path to the client certificate file.
    :type certfile: str | unicode
    :param keyfile: Full path to the client private key file.
    :type keyfile: str | unicode
    :param crlfile: Full path to the CRL file for validating certification
        expiry. This option is only available with Python 3.4+ or 2.7.9+.
    :type crlfile: str | unicode
    :param proc_ttl: The number of records read before the worker's process
        (multiprocessing pool of 1 process) is re-spawned. If set to ``0``
        or ``None``, the re-spawning is disabled. Default: ``5000``.
    :type proc_ttl: int
    """

    def __init__(self,
                 hosts='127.0.0.1:9092',
                 topic='default',
                 timeout=None,
                 callback=None,
                 job_size=1048576,
                 cafile=None,
                 certfile=None,
                 keyfile=None,
                 crlfile=None,
                 proc_ttl=5000):
        self._hosts = hosts
        self._topic = topic
        self._timeout = timeout
        self._callback = callback
        self._pool = None
        self._proc_ttl = proc_ttl
        self._logger = logging.getLogger('kq')
        self._consumer = kafka.KafkaConsumer(
            self._topic,
            group_id=self._topic,
            bootstrap_servers=self._hosts,
            max_partition_fetch_bytes=job_size * 2,
            ssl_cafile=cafile,
            ssl_certfile=certfile,
            ssl_keyfile=keyfile,
            ssl_crlfile=crlfile,
            consumer_timeout_ms=-1,
            enable_auto_commit=False,
            auto_offset_reset='latest',
        )

    def __del__(self):
        """Commit the Kafka consumer offsets and close the consumer."""
        if hasattr(self, '_consumer'):
            try:
                self._logger.info('Closing consumer ...')
                self._consumer.close()
            except Exception as e:  # pragma: no cover
                self._logger.warning('Failed to close consumer: {}'.format(e))

    def __repr__(self):
        """Return a string representation of the worker.

        :return: string representation of the worker
        :rtype: str | unicode
        """
        return 'Worker(topic={})'.format(self._topic)

    def _exec_callback(self, status, job, result, exception, traceback):
        """Execute the callback in a try-except block.

        :param status: The status of the job consumption. Possible values are
            ``timeout', ``failure`` and ``success``.
        :type status: str | unicode
        :param job: The job consumed by the worker
        :type job: kq.job.Job
        :param result: The result of the job execution.
        :type result: object
        :param exception: Exception raised while the job was running (i.e.
            status was ``failure``), or ``None`` if there were no errors
            (i.e. status was ``success``).
        :type exception: Exception | None
        :param traceback: The stacktrace of the exception (i.e. status was
            ``failure``) was running or ``None`` if there were no errors.
        :type traceback: str | unicode | None
        """
        if self._callback is not None:
            try:
                self._logger.info('Executing callback ...')
                self._callback(status, job, result, exception, traceback)
            except Exception as e:
                self._logger.exception('Callback failed: {}'.format(e))

    def _consume_record(self, record):
        """De-serialize the message and execute the incoming job.

        :param record: Record fetched from the Kafka topic.
        :type record: kafka.consumer.fetcher.ConsumerRecord
        """
        rec = rec_repr(record)
        self._logger.info('Processing {} ...'.format(rec))
        try:
            job = dill.loads(record.value)
        except:
            self._logger.warning('{} unloadable. Skipping ...'.format(rec))
        else:
            # Simple check for job validity
            if not (isinstance(job, Job)
                    and isinstance(job.args, collections.Iterable)
                    and isinstance(job.kwargs, collections.Mapping)
                    and callable(job.func)):
                self._logger.warning('{} malformed. Skipping ...'.format(rec))
                return
            func, args, kwargs = job.func, job.args, job.kwargs
            self._logger.info('Running Job {}: {} ...'.format(
                job.id, func_repr(func, args, kwargs)
            ))
            try:
                timeout = self._timeout or job.timeout
                if timeout is None:
                    res = func(*args, **kwargs)
                else:
                    run = self._pool.apply_async(func, args, kwargs)
                    res = run.get(timeout)
            except mp.TimeoutError:
                self._logger.error('Job {} timed out after {} seconds.'
                                   .format(job.id, job.timeout))
                self._exec_callback('timeout', job, None, None, None)
            except Exception as e:
                self._logger.exception('Job {} failed: {}'.format(job.id, e))
                self._exec_callback('failure', job, None, e, tb.format_exc())
            else:
                self._logger.info('Job {} returned: {}'.format(job.id, res))
                self._exec_callback('success', job, res, None, None)

    def _refresh_pool(self):
        """Terminate the previous process pool and initialize a new one."""
        self._logger.info('Refreshing process pool ...')
        try:
            self._pool.terminate()
        except Exception as e:  # pragma: no cover
            self._logger.exception('Failed to terminate pool: {}'.format(e))
        finally:
            self._pool = mp.Pool(processes=1)

    @property
    def consumer(self):
        """Return the Kafka consumer object.

        :return: Kafka consumer object.
        :rtype: kafka.consumer.KafkaConsumer
        """
        return self._consumer

    @property
    def hosts(self):
        """Return the list of Kafka host names and ports.

        :return: list of Kafka host names and ports
        :rtype: [str]
        """
        return self._hosts.split(',')

    @property
    def topic(self):
        """Return the name of the Kafka topic in use.

        :return: Name of the Kafka topic in use.
        :rtype: str | unicode
        """
        return self._topic

    @property
    def timeout(self):
        """Return the job timeout threshold in seconds.

        :return: Job timeout threshold in seconds.
        :rtype: int
        """
        return self._timeout

    def start(self):
        """Start fetching and processing enqueued jobs in the topic.

        Once started, the worker will continuously poll the Kafka broker in
        a loop until jobs are available in the topic partitions. The loop is
        only stopped via external triggers (e.g. keyboard interrupts).
        """
        self._logger.info('Starting {} ...'.format(self))
        self._pool = mp.Pool(processes=1)
        records_read = 0

        try:
            for record in self._consumer:
                self._consume_record(record)
                self._consumer.commit()
                if self._proc_ttl and records_read >= self._proc_ttl:
                    self._refresh_pool()
                    records_read = 0
                else:
                    records_read += 1

        except KeyboardInterrupt:  # pragma: no cover
            self._logger.info('Stopping {} ...'.format(self))
            self._pool.terminate()  # TODO not sure if necessary



from __future__ import absolute_import, print_function, unicode_literals


def rec_repr(record):
    """Return the string representation of the consumer record.

    :param record: Record fetched from the Kafka topic.
    :type record: kafka.consumer.fetcher.ConsumerRecord
    :return: String representation of the consumer record.
    :rtype: str | unicode
    """
    return 'Record(topic={}, partition={}, offset={})'.format(
        record.topic, record.partition, record.offset
    )


def func_repr(func, args, kwargs):
    """Return the string representation of the function call.

    :param func: Function to represent.
    :type func: callable
    :param args: Function arguments.
    :type args: list | collections.Iterable
    :param kwargs: Function keyword arguments.
    :type kwargs: dict | collections.Mapping
    :return: String representation of the function call.
    :rtype: str | unicode
    """
    params = list(map(repr, args))
    params.extend(k + '=' + repr(v) for k, v in sorted(kwargs.items()))
    return '{}.{}({})'.format(
        getattr(func, '__module__', '?'),
        getattr(func, '__name__', '?'),
        ', '.join(params)
    )



from __future__ import absolute_import, print_function, unicode_literals

import kafka


# TODO need to add more functionality
class Manager(object):
    """KQ manager.

    Here is an example of initializing and using a manager:

    .. code-block:: python

        from kq import Manager

        manager = Manager(
            hosts='host:7000,host:8000',
            cafile='/my/files/cafile',
            certfile='/my/files/certfile',
            keyfile='/my/files/keyfile',
            crlfile='/my/files/crlfile'
        )
        manager.info()

    .. note::

        There is not much happening with this class right now. It's still
        a work in progress.

    :param hosts: Comma-separated Kafka hostnames and ports. For example,
        ``"localhost:9000,localhost:8000,192.168.1.1:7000"`` is a valid
        input string. Default: ``"127.0.0.1:9092"``.
    :type hosts: str | unicode
    :param cafile: Full path to the trusted CA certificate file.
    :type cafile: str | unicode
    :param certfile: Full path to the client certificate file.
    :type certfile: str | unicode
    :param keyfile: Full path to the client private key file.
    :type keyfile: str | unicode
    :param crlfile: Full path to the CRL file for validating certification
        expiry. This option is only available with Python 3.4+ or 2.7.9+.
    :type crlfile: str | unicode
    """

    def __init__(self,
                 hosts='127.0.0.1:9092',
                 cafile=None,
                 certfile=None,
                 keyfile=None,
                 crlfile=None):
        self._hosts = hosts
        self._consumer = kafka.KafkaConsumer(
            bootstrap_servers=self._hosts,
            ssl_cafile=cafile,
            ssl_certfile=certfile,
            ssl_keyfile=keyfile,
            ssl_crlfile=crlfile,
            consumer_timeout_ms=-1,
            enable_auto_commit=True,
            auto_offset_reset='latest',
        )

    def __repr__(self):
        """Return a string representation of the queue.

        :return: string representation of the queue
        :rtype: str | unicode
        """
        return 'Manager(hosts={})'.format(self._hosts)

    @property
    def consumer(self):
        """Return the Kafka consumer object.

        :return: Kafka consumer object.
        :rtype: kafka.consumer.KafkaConsumer
        """
        return self._consumer

    @property
    def hosts(self):
        """Return the list of Kafka host names and ports.

        :return: list of Kafka host names and ports
        :rtype: [str]
        """
        return self._hosts.split(',')

    def info(self):
        """Print the offset information for all topics and partitions."""
        print('Offsets per Topic:')
        for topic in self._consumer.topics():
            print('\nTopic {}:\n'.format(topic))
            partitions = self._consumer.partitions_for_topic(topic)
            if partitions is None:  # pragma: no cover
                print('    Polling failed (please try again)')
                continue
            for partition in self._consumer.partitions_for_topic(topic):
                topic_partition = kafka.TopicPartition(topic, partition)
                self._consumer.assign([topic_partition])
                offset = self._consumer.position(topic_partition)
                print('    Partition {:<3}: {}'.format(partition, offset))



from __future__ import absolute_import, print_function, unicode_literals

import logging
import sys

import mock
import pytest

from kq import cli, VERSION

from .utils import CaptureOutput

patch_object = getattr(mock.patch, 'object')


@pytest.fixture(autouse=True)
def manager(monkeypatch):
    mock_manager_inst = mock.MagicMock()
    mock_manager_cls = mock.MagicMock(return_value=mock_manager_inst)
    monkeypatch.setattr('kq.Manager', mock_manager_cls)
    return mock_manager_cls, mock_manager_inst


@pytest.fixture(autouse=True)
def worker(monkeypatch):
    mock_worker_inst = mock.MagicMock()
    mock_worker_cls = mock.MagicMock(return_value=mock_worker_inst)
    monkeypatch.setattr('kq.Worker', mock_worker_cls)
    return mock_worker_cls, mock_worker_inst


@pytest.fixture(autouse=True)
def logger(monkeypatch):
    mock_logger_inst = mock.MagicMock()
    mock_get_logger = mock.MagicMock()
    mock_get_logger.return_value = mock_logger_inst
    monkeypatch.setattr('logging.getLogger', mock_get_logger)
    return mock_logger_inst


def test_help_menu():
    with patch_object(sys, 'argv', ['kq', '--help']):
        with CaptureOutput() as output, pytest.raises(SystemExit):
            cli.entry_point()
        assert len(output) == 1
        assert 'Usage' in output[0]
        assert 'Options' in output[0]


def test_version():
    with patch_object(sys, 'argv', ['kq', '--version']):
        with CaptureOutput() as output, pytest.raises(SystemExit):
            cli.entry_point()
        assert len(output) == 1
        assert output[0] == VERSION + '\n'


def test_info(manager, logger):
    manager_cls, manager_inst = manager

    test_arguments = [
        'kq',
        'info',
        '--hosts=host:6000,host:7000',
        '--cafile=/test/files/cafile',
        '--certfile=/test/files/certfile',
        '--keyfile=/test/files/keyfile',
        '--crlfile=/test/files/crlfile'
    ]
    with patch_object(sys, 'argv', test_arguments):
        cli.entry_point()

    logger.setLevel.assert_called_once_with(logging.WARNING)
    manager_cls.assert_called_once_with(
        hosts='host:6000,host:7000',
        cafile='/test/files/cafile',
        certfile='/test/files/certfile',
        keyfile='/test/files/keyfile',
        crlfile='/test/files/crlfile'
    )
    manager_inst.info.assert_called_once()


def test_worker(worker, logger):
    worker_cls, worker_inst = worker

    test_arguments = [
        'kq',
        'worker',
        '--hosts=host:6000,host:7000',
        '--topic=foo',
        '--timeout=4000',
        '--job-size=3000000',
        '--cafile=/test/files/cafile',
        '--certfile=/test/files/certfile',
        '--keyfile=/test/files/keyfile',
        '--crlfile=/test/files/crlfile',
        '--proc-ttl=1000'
    ]
    with patch_object(sys, 'argv', test_arguments):
        cli.entry_point()

    logger.setLevel.assert_called_once_with(logging.WARNING)
    worker_cls.assert_called_once_with(
        hosts='host:6000,host:7000',
        topic='foo',
        timeout=4000,
        callback=None,
        job_size=3000000,
        cafile='/test/files/cafile',
        certfile='/test/files/certfile',
        keyfile='/test/files/keyfile',
        crlfile='/test/files/crlfile',
        proc_ttl=1000
    )
    worker_inst.start.assert_called_once()


def test_callback(worker, logger):
    worker_cls, worker_inst = worker

    test_arguments = [
        'kq',
        'worker',
        '--hosts=host:6000,host:7000',
        '--topic=foo',
        '--timeout=4000',
        '--callback=/invalid/path'
    ]
    with patch_object(sys, 'argv', test_arguments):
        cli.entry_point()

    logger.exception.assert_called_once()
    logger.setLevel.assert_called_once_with(logging.WARNING)
    worker_cls.assert_called_with(
        hosts='host:6000,host:7000',
        topic='foo',
        timeout=4000,
        callback=None,
        job_size=1048576,
        cafile=None,
        certfile=None,
        keyfile=None,
        crlfile=None,
        proc_ttl=5000
    )
    worker_inst.start.assert_called_once()


def test_verbose(worker, logger):
    worker_cls, worker_inst = worker

    with patch_object(sys, 'argv', ['kq', 'worker', '--verbose']):
        cli.entry_point()

    logger.setLevel.assert_called_once_with(logging.DEBUG)
    worker_cls.assert_called_with(
        hosts='127.0.0.1',
        topic='default',
        timeout=None,
        callback=None,
        job_size=1048576,
        cafile=None,
        certfile=None,
        keyfile=None,
        crlfile=None,
        proc_ttl=5000
    )
    worker_inst.start.assert_called_once()



from __future__ import absolute_import, print_function, unicode_literals

from collections import namedtuple

import dill
import mock
import pytest

from kq import Job, Worker

from .utils import (
    success_func,
    failure_func,
    timeout_func
)

MockRecord = namedtuple(
    'MockRecord',
    ['topic', 'partition', 'offset', 'value']
)
# Mocks for patching KafkaConsumer
mock_consumer = mock.MagicMock()
mock_consumer.topics.return_value = ['foo', 'bar', 'baz']
mock_consumer.partitions_for_topic.return_value = [1, 2, 3]
mock_consumer.position.return_value = 100
mock_consumer_cls = mock.MagicMock()
mock_consumer_cls.return_value = mock_consumer

# Mocks for patching the logging module
success_job = Job(
    id='100',
    timestamp=1,
    topic='foo',
    func=success_func,
    args=[1, 2],
    kwargs={'c': 3},
    timeout=None,
)
failure_job = Job(
    id='200',
    timestamp=2,
    topic='foo',
    func=failure_func,
    args=[1, 2, 3],
    kwargs={},
    timeout=100,
)
timeout_job = Job(
    id='300',
    timestamp=3,
    topic='foo',
    func=timeout_func,
    args=[2, 3, 4],
    kwargs={},
    timeout=100,
)
value1 = dill.dumps(success_job)
value2 = dill.dumps(failure_job)
value3 = dill.dumps(timeout_job)
value4 = 'This is an unpicklable value'
value5 = dill.dumps(['This is not a job!'])

# Mocks for consumer records
rec11 = MockRecord(topic='foo', partition=1, offset=1, value=value1)
rec11_repr = 'Record(topic=foo, partition=1, offset=1)'
rec12 = MockRecord(topic='foo', partition=1, offset=2, value=value2)
rec12_repr = 'Record(topic=foo, partition=1, offset=2)'
rec21 = MockRecord(topic='foo', partition=2, offset=1, value=value3)
rec21_repr = 'Record(topic=foo, partition=2, offset=1)'
rec22 = MockRecord(topic='foo', partition=2, offset=2, value=value4)
rec22_repr = 'Record(topic=foo, partition=2, offset=2)'
rec34 = MockRecord(topic='foo', partition=3, offset=4, value=value5)
rec34_repr = 'Record(topic=foo, partition=3, offset=4)'


@pytest.fixture(autouse=True)
def consumer(monkeypatch):
    monkeypatch.setattr('kafka.KafkaConsumer', mock_consumer_cls)
    mock_consumer.reset_mock()


@pytest.fixture(autouse=True)
def logger(monkeypatch):
    mock_logger = mock.MagicMock()
    mock_get_logger = mock.MagicMock()
    mock_get_logger.return_value = mock_logger
    monkeypatch.setattr('logging.getLogger', mock_get_logger)
    return mock_logger


@pytest.fixture(autouse=True)
def callback():
    return mock.MagicMock()


def test_init(logger, callback):
    worker = Worker(
        hosts='host:7000,host:8000',
        topic='foo',
        timeout=1000,
        callback=callback,
        job_size=10000000,
        cafile='/test/files/cafile',
        certfile='/test/files/certfile',
        keyfile='/test/files/keyfile',
        crlfile='/test/files/crlfile'
    )
    mock_consumer_cls.assert_called_once_with(
        'foo',
        group_id='foo',
        bootstrap_servers='host:7000,host:8000',
        max_partition_fetch_bytes=20000000,
        ssl_cafile='/test/files/cafile',
        ssl_certfile='/test/files/certfile',
        ssl_keyfile='/test/files/keyfile',
        ssl_crlfile='/test/files/crlfile',
        consumer_timeout_ms=-1,
        enable_auto_commit=False,
        auto_offset_reset='latest',
    )
    assert repr(worker) == 'Worker(topic=foo)'
    assert worker.hosts == ['host:7000', 'host:8000']
    assert worker.timeout == 1000
    assert worker.topic == 'foo'
    assert worker.consumer == mock_consumer
    assert not callback.called
    assert not logger.info.called


def test_start_job_success(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec11])
    worker = Worker(
        hosts='localhost',
        topic='foo',
        callback=callback,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec11_repr)),
        mock.call('Running Job 100: tests.utils.success_func(1, 2, c=3) ...'),
        mock.call('Job 100 returned: (1, 2, 3)'),
        mock.call('Executing callback ...')
    ])
    callback.assert_called_once_with(
        'success', success_job, (1, 2, 3), None, None
    )


def test_start_job_failure(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec12])
    worker = Worker(
        hosts='localhost',
        topic='foo',
        timeout=1000,
        callback=callback,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec12_repr)),
        mock.call('Running Job 200: tests.utils.failure_func(1, 2, 3) ...'),
        mock.call('Executing callback ...')
    ])
    logger.exception.assert_called_with('Job 200 failed: failed!')
    assert len(callback.call_args_list) == 1

    callback_args = callback.call_args_list[0][0]
    assert callback_args[0] == 'failure'
    assert callback_args[1] == failure_job
    assert callback_args[2] is None
    assert isinstance(callback_args[3], ValueError)
    assert isinstance(callback_args[4], str)


def test_start_job_timeout(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec21])
    worker = Worker(
        hosts='localhost',
        topic='foo',
        timeout=1000,
        callback=callback,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec21_repr)),
        mock.call('Running Job 300: tests.utils.timeout_func(2, 3, 4) ...'),
        mock.call('Executing callback ...')
    ])
    logger.error.assert_called_once_with(
        'Job 300 timed out after 100 seconds.'
    )
    callback.assert_called_once_with(
        'timeout', timeout_job, None, None, None
    )


def test_start_job_unloadable(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec22])
    worker = Worker(
        hosts='localhost',
        topic='foo',
        timeout=1000,
        callback=callback,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec22_repr)),
    ])
    logger.warning.assert_called_once_with(
        '{} unloadable. Skipping ...'.format(rec22_repr)
    )
    assert not callback.called


def test_start_job_malformed(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec34])
    worker = Worker(
        hosts='localhost',
        topic='foo',
        timeout=1000,
        callback=callback,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec34_repr)),
    ])
    logger.warning.assert_called_once_with(
        '{} malformed. Skipping ...'.format(rec34_repr)
    )
    assert not callback.called


def test_start_job_callback_fail(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec11])
    expected_error = KeyError('foo')
    callback.side_effect = expected_error
    worker = Worker(
        hosts='localhost',
        topic='foo',
        callback=callback,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec11_repr)),
        mock.call('Running Job 100: tests.utils.success_func(1, 2, c=3) ...'),
        mock.call('Job 100 returned: (1, 2, 3)'),
        mock.call('Executing callback ...')
    ])
    logger.exception.assert_called_once_with(
        'Callback failed: {}'.format(expected_error)
    )


def test_start_proc_ttl_reached(logger, callback):
    mock_consumer.__iter__ = lambda x: iter([rec11, rec11])
    worker = Worker(
        hosts='localhost',
        topic='foo',
        callback=callback,
        proc_ttl=1,
    )
    worker.start()
    logger.info.assert_has_calls([
        mock.call('Starting Worker(topic=foo) ...'),
        mock.call('Processing {} ...'.format(rec11_repr)),
        mock.call('Running Job 100: tests.utils.success_func(1, 2, c=3) ...'),
        mock.call('Job 100 returned: (1, 2, 3)'),
        mock.call('Executing callback ...'),
        mock.call('Processing {} ...'.format(rec11_repr)),
        mock.call('Running Job 100: tests.utils.success_func(1, 2, c=3) ...'),
        mock.call('Job 100 returned: (1, 2, 3)'),
        mock.call('Executing callback ...'),
        mock.call('Refreshing process pool ...'),
    ])
    callback.assert_called_with(
        'success', success_job, (1, 2, 3), None, None
    )


from __future__ import absolute_import, print_function, unicode_literals

import logging

from kq.job import Job
from kq.manager import Manager
from kq.worker import Worker
from kq.queue import Queue
from kq.version import VERSION

__all__ = ['Job', 'Manager', 'Worker', 'Queue', 'VERSION']

# Reduce logging noise from PyKafka
for name, logger in logging.Logger.manager.loggerDict.items():
    if name.startswith('kafka') and isinstance(logger, logging.Logger):
        logger.setLevel(logging.CRITICAL)



from setuptools import setup, find_packages

version = {}
with open('./kq/version.py') as fp:
    exec(fp.read(), version)

setup(
    name='kq',
    description='Kafka Job Queue for Python',
    version=version['VERSION'],
    author='Joohwan Oh',
    author_email='joohwan.oh@outlook.com',
    url='https://github.com/joowani/kq',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'dill>=0.2.5',
        'docopt>=0.6.2',
        'kafka-python>=1.3.1'
    ],
    entry_points={
        'console_scripts': [
            'kq = kq.cli:entry_point',
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
        'Topic :: System :: Monitoring',
    ]
)



"""
KQ (Kafka Queue) command line tool

Usage:
  kq (worker|info) [--hosts=<hosts>]
                   [--topic=<topic>]
                   [--timeout=<timeout>]
                   [--callback=<callback>]
                   [--job-size=<job-size>]
                   [--cafile=<cafile>]
                   [--certfile=<certfile>]
                   [--keyfile=<keyfile>]
                   [--crlfile=<crlfile>]
                   [--proc-ttl=<proc-ttl>]
                   [--verbose]
  kq --help
  kq --version

Options:
  --hosts=<hosts>        Comma-separated Kafka hosts [default: 127.0.0.1]
  --topic=<topic>        Name of the Kafka topic [default: default]
  --job-size=<job-size>  Maximum job size in bytes [default: 1048576]
  --timeout=<timeout>    Job timeout threshold in seconds
  --callback=<callback>  Python module containing the callback function
  --cafile=<cafile>      Full path to SSL trusted CA certificate
  --certfile=<certfile>  Full path to SSL client certificate
  --keyfile=<keyfile>    Full path to SSL private key
  --crlfile=<crlfile>    Full path to SSL crlfile for verifying expiry
  --proc-ttl=<proc-ttl>  Records read before re-spawning process [default: 5000]
  --verbose              Turn on debug logging output
  --help                 Display this help menu
  --version              Display the version of KQ

"""

import os
import sys
import logging

import docopt

import kq


def entry_point():
    args = docopt.docopt(__doc__, version=kq.VERSION)

    # Set up the logger
    logger = logging.getLogger('kq')
    if args['--verbose']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.WARNING)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(logging.Formatter(
        '[%(asctime)s] [%(levelname)s] %(message)s',
    ))
    logger.addHandler(stream_handler)

    # Import the callback function if given
    callback = None
    if args['--callback']:  # pragma: no cover
        try:
            path = os.path.abspath(
                os.path.expanduser(args['--callback'])
            )
            if sys.version_info.major == 2:
                import imp
                module = imp.load_source('kqconfig', path)
            elif sys.version_info.minor < 5:
                from importlib.machinery import SourceFileLoader as Loader
                module = Loader('kqconfig', path).load_module()
            else:
                from importlib import util
                spec = util.spec_from_file_location("kqconfig", path)
                module = util.module_from_spec(spec)
                spec.loader.exec_module(module)
            callback = getattr(module, 'callback')
        except Exception as e:
            logger.exception('Failed to load callback: {}'.format(e))

    if args['info']:
        kq.Manager(
            hosts=args['--hosts'],
            cafile=args['--cafile'],
            certfile=args['--certfile'],
            keyfile=args['--keyfile'],
            crlfile=args['--crlfile'],
        ).info()

    elif args['worker']:
        timeout = args['--timeout']
        kq.Worker(
            hosts=args['--hosts'],
            topic=args['--topic'],
            timeout=int(timeout) if timeout else None,
            callback=callback,
            job_size=int(args['--job-size']),
            cafile=args['--cafile'],
            certfile=args['--certfile'],
            keyfile=args['--keyfile'],
            crlfile=args['--crlfile'],
            proc_ttl=int(args['--proc-ttl'])
        ).start()
